from __future__ import annotations

import functools
import inspect
from typing import Any

from coredis.config import Config
from coredis.exceptions import CommandSyntaxError
from coredis.typing import (
    Callable,
    Coroutine,
    Iterable,
    Optional,
    ParamSpec,
    Set,
    TypeVar,
    Union,
)

R = TypeVar("R")
P = ParamSpec("P")


class MutuallyExclusiveParametersError(CommandSyntaxError):
    def __init__(self, arguments: Set[str], details: Optional[str]):
        message = (
            f"The [{','.join(arguments)}] parameters are mutually exclusive."
            f"{' '+details if details else ''}"
        )
        super().__init__(arguments, message)


class MutuallyInclusiveParametersMissing(CommandSyntaxError):
    def __init__(self, arguments: Set[str], leaders: Set[str], details: Optional[str]):
        if leaders:
            message = (
                f"The [{','.join(arguments)}] parameters(s)"
                " must be provided together with [{','.join(leaders)}]."
                f"{' ' + details if details else ''}"
            )
        else:
            message = (
                f"The [{','.join(arguments)}] parameters are mutually"
                " inclusive and must be provided together."
                f"{' '+details if details else ''}"
            )
        super().__init__(arguments, message)


def mutually_exclusive_parameters(
    *exclusive_params: Union[str, Iterable[str]], details: Optional[str] = None
) -> Callable[
    [Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]
]:
    primary = {k for k in exclusive_params if isinstance(k, str)}
    secondary = [k for k in set(exclusive_params) - primary]

    def wrapper(
        func: Callable[P, Coroutine[Any, Any, R]]
    ) -> Callable[P, Coroutine[Any, Any, R]]:
        sig = inspect.signature(func)

        @functools.wraps(func)
        async def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            if not Config.optimized:
                call_args = sig.bind_partial(*args, **kwargs)
                params = {
                    k
                    for k in primary
                    if not call_args.arguments.get(k)
                    == getattr(sig.parameters.get(k), "default")
                }

                if params:
                    for group in secondary:
                        for k in group:
                            if not call_args.arguments.get(k) == getattr(
                                sig.parameters.get(k), "default"
                            ):
                                params.add(k)

                                break

                if len(params) > 1:
                    raise MutuallyExclusiveParametersError(params, details)

            return await func(*args, **kwargs)

        return wrapped if not Config.optimized else func

    return wrapper


def mutually_inclusive_parameters(
    *inclusive_params: str,
    leaders: Optional[Iterable[str]] = None,
    details: Optional[str] = None,
) -> Callable[
    [Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]
]:
    _leaders = set(leaders or [])
    _inclusive_params = set(inclusive_params)

    def wrapper(
        func: Callable[P, Coroutine[Any, Any, R]]
    ) -> Callable[P, Coroutine[Any, Any, R]]:
        sig = inspect.signature(func)

        @functools.wraps(func)
        async def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            if not Config.optimized:
                call_args = sig.bind_partial(*args, **kwargs)
                params = {
                    k
                    for k in _inclusive_params | _leaders
                    if not call_args.arguments.get(k)
                    == getattr(sig.parameters.get(k), "default")
                }
                if _leaders and _leaders & params != _leaders and len(params) > 0:
                    raise MutuallyInclusiveParametersMissing(
                        _inclusive_params, _leaders, details
                    )
                elif not _leaders and params and len(params) != len(_inclusive_params):
                    raise MutuallyInclusiveParametersMissing(
                        _inclusive_params, _leaders, details
                    )
            return await func(*args, **kwargs)

        return wrapped if not Config.optimized else func

    return wrapper


def ensure_iterable_valid(
    argument: str,
) -> Callable[
    [Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]
]:
    def iterable_valid(value: Any) -> bool:
        return isinstance(value, Iterable) and not isinstance(value, (str, bytes))

    def wrapper(
        func: Callable[P, Coroutine[Any, Any, R]]
    ) -> Callable[P, Coroutine[Any, Any, R]]:
        sig = inspect.signature(func)
        expected_type = sig.parameters[argument].annotation

        @functools.wraps(func)
        async def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            if Config.optimized:
                return await func(*args, **kwargs)
            bound = sig.bind_partial(*args, **kwargs)
            value = bound.arguments.get(argument)
            if not iterable_valid(value):
                raise TypeError(
                    f"{func.__name__} parameter {argument}={value!r} "
                    f"violates expected iterable of type {expected_type}"
                )

            return await func(*args, **kwargs)

        return wrapped if not Config.optimized else func

    return wrapper



from __future__ import annotations

from coredis._utils import CaseAndEncodingInsensitiveEnum


class PureToken(CaseAndEncodingInsensitiveEnum):
    """
    Enum for using pure-tokens with the redis api.
    """

    #: Used by:
    #:
    #:  - ``GRAPH.CONSTRAINT CREATE``
    #:  - ``GRAPH.CONTRAINT DROP``
    UNIQUE = b"UNIQUE"

    #: Used by:
    #:
    #:  - ``GRAPH.CONSTRAINT CREATE``
    #:  - ``GRAPH.CONTRAINT DROP``
    MANDATORY = b"MANDATORY"

    #: Used by:
    #:
    #:  - ``ACL LOG``
    RESET = b"RESET"

    #: Used by:
    #:
    #:  - ``BGSAVE``
    SCHEDULE = b"SCHEDULE"

    #: Used by:
    #:
    #:  - ``BITCOUNT``
    #:  - ``BITPOS``
    BIT = b"BIT"

    #: Used by:
    #:
    #:  - ``BITCOUNT``
    #:  - ``BITPOS``
    BYTE = b"BYTE"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    FAIL = b"FAIL"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    SAT = b"SAT"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    WRAP = b"WRAP"

    #: Used by:
    #:
    #:  - ``BITOP``
    AND = b"AND"

    #: Used by:
    #:
    #:  - ``BITOP``
    NOT = b"NOT"

    #: Used by:
    #:
    #:  - ``BITOP``
    OR = b"OR"

    #: Used by:
    #:
    #:  - ``BITOP``
    XOR = b"XOR"

    #: Used by:
    #:
    #:  - ``BLMOVE``
    #:  - ``BLMPOP``
    #:  - ``LMOVE``
    #:  - ``LMPOP``
    LEFT = b"LEFT"

    #: Used by:
    #:
    #:  - ``BLMOVE``
    #:  - ``BLMPOP``
    #:  - ``LMOVE``
    #:  - ``LMPOP``
    RIGHT = b"RIGHT"

    #: Used by:
    #:
    #:  - ``BZMPOP``
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    #:  - ``ZINTER``
    #:  - ``ZINTERSTORE``
    #:  - ``ZMPOP``
    #:  - ``ZUNION``
    #:  - ``ZUNIONSTORE``
    MAX = b"MAX"

    #: Used by:
    #:
    #:  - ``BZMPOP``
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    #:  - ``ZINTER``
    #:  - ``ZINTERSTORE``
    #:  - ``ZMPOP``
    #:  - ``ZUNION``
    #:  - ``ZUNIONSTORE``
    MIN = b"MIN"

    #: Used by:
    #:
    #:  - ``CLIENT CACHING``
    #:  - ``CLIENT KILL``
    #:  - ``SCRIPT DEBUG``
    NO = b"NO"

    #: Used by:
    #:
    #:  - ``CLIENT CACHING``
    #:  - ``CLIENT KILL``
    #:  - ``SCRIPT DEBUG``
    YES = b"YES"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    #:  - ``CLIENT LIST``
    MASTER = b"MASTER"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    #:  - ``CLIENT LIST``
    NORMAL = b"NORMAL"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    #:  - ``CLIENT LIST``
    PUBSUB = b"PUBSUB"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    #:  - ``CLIENT LIST``
    REPLICA = b"REPLICA"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    SLAVE = b"SLAVE"

    #: Used by:
    #:
    #:  - ``CLIENT NO-EVICT``
    #:  - ``CLIENT NO-TOUCH``
    #:  - ``CLIENT REPLY``
    #:  - ``CLIENT TRACKING``
    OFF = b"OFF"

    #: Used by:
    #:
    #:  - ``CLIENT NO-EVICT``
    #:  - ``CLIENT NO-TOUCH``
    #:  - ``CLIENT REPLY``
    #:  - ``CLIENT TRACKING``
    ON = b"ON"

    #: Used by:
    #:
    #:  - ``CLIENT PAUSE``
    ALL = b"ALL"

    #: Used by:
    #:
    #:  - ``CLIENT PAUSE``
    WRITE = b"WRITE"

    #: Used by:
    #:
    #:  - ``CLIENT REPLY``
    SKIP = b"SKIP"

    #: Used by:
    #:
    #:  - ``CLIENT TRACKING``
    BCAST = b"BCAST"

    #: Used by:
    #:
    #:  - ``CLIENT TRACKING``
    NOLOOP = b"NOLOOP"

    #: Used by:
    #:
    #:  - ``CLIENT TRACKING``
    OPTIN = b"OPTIN"

    #: Used by:
    #:
    #:  - ``CLIENT TRACKING``
    OPTOUT = b"OPTOUT"

    #: Used by:
    #:
    #:  - ``CLIENT UNBLOCK``
    ERROR = b"ERROR"

    #: Used by:
    #:
    #:  - ``CLIENT UNBLOCK``
    TIMEOUT = b"TIMEOUT"

    #: Used by:
    #:
    #:  - ``CLUSTER FAILOVER``
    #:  - ``FAILOVER``
    #:  - ``SHUTDOWN``
    #:  - ``XCLAIM``
    FORCE = b"FORCE"

    #: Used by:
    #:
    #:  - ``CLUSTER FAILOVER``
    TAKEOVER = b"TAKEOVER"

    #: Used by:
    #:
    #:  - ``CLUSTER RESET``
    HARD = b"HARD"

    #: Used by:
    #:
    #:  - ``CLUSTER RESET``
    SOFT = b"SOFT"

    #: Used by:
    #:
    #:  - ``CLUSTER SETSLOT``
    STABLE = b"STABLE"

    #: Used by:
    #:
    #:  - ``COPY``
    #:  - ``FUNCTION LOAD``
    #:  - ``FUNCTION RESTORE``
    #:  - ``MIGRATE``
    #:  - ``RESTORE``
    #:  - ``RESTORE-ASKING``
    REPLACE = b"REPLACE"

    #: Used by:
    #:
    #:  - ``EXPIRE``
    #:  - ``EXPIREAT``
    #:  - ``PEXPIRE``
    #:  - ``PEXPIREAT``
    #:  - ``ZADD``
    GT = b"GT"

    #: Used by:
    #:
    #:  - ``EXPIRE``
    #:  - ``EXPIREAT``
    #:  - ``PEXPIRE``
    #:  - ``PEXPIREAT``
    #:  - ``ZADD``
    LT = b"LT"

    #: Used by:
    #:
    #:  - ``EXPIRE``
    #:  - ``EXPIREAT``
    #:  - ``GEOADD``
    #:  - ``JSON.SET``
    #:  - ``PEXPIRE``
    #:  - ``PEXPIREAT``
    #:  - ``SET``
    #:  - ``ZADD``
    NX = b"NX"

    #: Used by:
    #:
    #:  - ``EXPIRE``
    #:  - ``EXPIREAT``
    #:  - ``GEOADD``
    #:  - ``JSON.SET``
    #:  - ``PEXPIRE``
    #:  - ``PEXPIREAT``
    #:  - ``SET``
    #:  - ``ZADD``
    XX = b"XX"

    #: Used by:
    #:
    #:  - ``FAILOVER``
    #:  - ``SHUTDOWN``
    ABORT = b"ABORT"

    #: Used by:
    #:
    #:  - ``FLUSHALL``
    #:  - ``FLUSHDB``
    #:  - ``FUNCTION FLUSH``
    #:  - ``SCRIPT FLUSH``
    ASYNC = b"ASYNC"

    #: Used by:
    #:
    #:  - ``FLUSHALL``
    #:  - ``FLUSHDB``
    #:  - ``FUNCTION FLUSH``
    #:  - ``SCRIPT DEBUG``
    #:  - ``SCRIPT FLUSH``
    SYNC = b"SYNC"

    #: Used by:
    #:
    #:  - ``FUNCTION LIST``
    WITHCODE = b"WITHCODE"

    #: Used by:
    #:
    #:  - ``FUNCTION RESTORE``
    APPEND = b"APPEND"

    #: Used by:
    #:
    #:  - ``FUNCTION RESTORE``
    FLUSH = b"FLUSH"

    #: Used by:
    #:
    #:  - ``GEOADD``
    #:  - ``ZADD``
    CHANGE = b"CH"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``GEODIST``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    FT = b"FT"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``GEODIST``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    KM = b"KM"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``GEODIST``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    M = b"M"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``GEODIST``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    MI = b"MI"

    #: Used by:
    #:
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    ANY = b"ANY"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.SEARCH``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    #:  - ``SORT``
    #:  - ``SORT_RO``
    ASC = b"ASC"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.SEARCH``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    #:  - ``SORT``
    #:  - ``SORT_RO``
    DESC = b"DESC"

    #: Used by:
    #:
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    WITHCOORD = b"WITHCOORD"

    #: Used by:
    #:
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    WITHDIST = b"WITHDIST"

    #: Used by:
    #:
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    WITHHASH = b"WITHHASH"

    #: Used by:
    #:
    #:  - ``GEOSEARCHSTORE``
    STOREDIST = b"STOREDIST"

    #: Used by:
    #:
    #:  - ``GETEX``
    PERSIST = b"PERSIST"

    #: Used by:
    #:
    #:  - ``HRANDFIELD``
    WITHVALUES = b"WITHVALUES"

    #: Used by:
    #:
    #:  - ``LCS``
    IDX = b"IDX"

    #: Used by:
    #:
    #:  - ``LCS``
    LEN = b"LEN"

    #: Used by:
    #:
    #:  - ``LCS``
    WITHMATCHLEN = b"WITHMATCHLEN"

    #: Used by:
    #:
    #:  - ``LINSERT``
    AFTER = b"AFTER"

    #: Used by:
    #:
    #:  - ``LINSERT``
    BEFORE = b"BEFORE"

    #: Used by:
    #:
    #:  - ``MIGRATE``
    COPY = b"COPY"

    #: Used by:
    #:
    #:  - ``MIGRATE``
    EMPTY_STRING = b""

    #: Used by:
    #:
    #:  - ``RESTORE``
    #:  - ``RESTORE-ASKING``
    ABSTTL = b"ABSTTL"

    #: Used by:
    #:
    #:  - ``SET``
    GET = b"GET"

    #: Used by:
    #:
    #:  - ``SET``
    KEEPTTL = b"KEEPTTL"

    #: Used by:
    #:
    #:  - ``SHUTDOWN``
    NOSAVE = b"NOSAVE"

    #: Used by:
    #:
    #:  - ``SHUTDOWN``
    NOW = b"NOW"

    #: Used by:
    #:
    #:  - ``SHUTDOWN``
    SAVE = b"SAVE"

    #: Used by:
    #:
    #:  - ``SORT``
    #:  - ``SORT_RO``
    SORTING = b"ALPHA"

    #: Used by:
    #:
    #:  - ``XADD``
    #:  - ``XTRIM``
    APPROXIMATELY = b"~"

    #: Used by:
    #:
    #:  - ``XADD``
    AUTO_ID = b"*"

    #: Used by:
    #:
    #:  - ``XADD``
    #:  - ``XTRIM``
    EQUAL = b"="

    #: Used by:
    #:
    #:  - ``XADD``
    #:  - ``XTRIM``
    MAXLEN = b"MAXLEN"

    #: Used by:
    #:
    #:  - ``XADD``
    #:  - ``XTRIM``
    MINID = b"MINID"

    #: Used by:
    #:
    #:  - ``XADD``
    NOMKSTREAM = b"NOMKSTREAM"

    #: Used by:
    #:
    #:  - ``XAUTOCLAIM``
    #:  - ``XCLAIM``
    JUSTID = b"JUSTID"

    #: Used by:
    #:
    #:  - ``XGROUP CREATE``
    MKSTREAM = b"MKSTREAM"

    #: Used by:
    #:
    #:  - ``XGROUP CREATE``
    #:  - ``XGROUP SETID``
    NEW_ID = b"$"

    #: Used by:
    #:
    #:  - ``XINFO STREAM``
    FULL = b"FULL"

    #: Used by:
    #:
    #:  - ``XREADGROUP``
    NOACK = b"NOACK"

    #: Used by:
    #:
    #:  - ``ZADD``
    INCREMENT = b"INCR"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``FT.SUGGET``
    #:  - ``ZDIFF``
    #:  - ``ZINTER``
    #:  - ``ZRANDMEMBER``
    #:  - ``ZRANGE``
    #:  - ``ZRANGEBYSCORE``
    #:  - ``ZREVRANGE``
    #:  - ``ZREVRANGEBYSCORE``
    #:  - ``ZUNION``
    WITHSCORES = b"WITHSCORES"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    #:  - ``ZINTER``
    #:  - ``ZINTERSTORE``
    #:  - ``ZUNION``
    #:  - ``ZUNIONSTORE``
    SUM = b"SUM"

    #: Used by:
    #:
    #:  - ``ZRANGE``
    #:  - ``ZRANGESTORE``
    BYLEX = b"BYLEX"

    #: Used by:
    #:
    #:  - ``ZRANGE``
    #:  - ``ZRANGESTORE``
    BYSCORE = b"BYSCORE"

    #: Used by:
    #:
    #:  - ``ZRANGE``
    #:  - ``ZRANGESTORE``
    REV = b"REV"

    #: Used by:
    #:
    #:  - ``ZRANK``
    #:  - ``ZREVRANK``
    WITHSCORE = b"WITHSCORE"

    #: Used by:
    #:
    #:  - ``BF.INSERT``
    #:  - ``BF.RESERVE``
    NONSCALING = b"NONSCALING"

    #: Used by:
    #:
    #:  - ``BF.INFO``
    #:  - ``BF.INSERT``
    #:  - ``CF.INSERT``
    #:  - ``CF.INSERTNX``
    ITEMS = b"ITEMS"

    #: Used by:
    #:
    #:  - ``BF.INSERT``
    #:  - ``CF.INSERT``
    #:  - ``CF.INSERTNX``
    NOCREATE = b"NOCREATE"

    #: Used by:
    #:
    #:  - ``BF.INFO``
    CAPACITY = b"CAPACITY"

    #: Used by:
    #:
    #:  - ``BF.INFO``
    EXPANSION = b"EXPANSION"

    #: Used by:
    #:
    #:  - ``BF.INFO``
    FILTERS = b"FILTERS"

    #: Used by:
    #:
    #:  - ``BF.INFO``
    SIZE = b"SIZE"

    #: Used by:
    #:
    #:  - ``CMS.MERGE``
    WEIGHTS = b"WEIGHTS"

    #: Used by:
    #:
    #:  - ``TOPK.LIST``
    WITHCOUNT = b"WITHCOUNT"

    #: Used by:
    #:
    #:  - ``TDIGEST.MERGE``
    COMPRESSION = b"COMPRESSION"

    #: Used by:
    #:
    #:  - ``TDIGEST.MERGE``
    OVERRIDE = b"OVERRIDE"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    BLOCK = b"BLOCK"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.CREATE``
    COMPRESSED = b"COMPRESSED"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    FIRST = b"FIRST"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    LAST = b"LAST"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.CREATE``
    #:  - ``TS.DECRBY``
    #:  - ``TS.INCRBY``
    UNCOMPRESSED = b"UNCOMPRESSED"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    AVG = b"AVG"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    COUNT = b"COUNT"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    RANGE = b"RANGE"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    STD_P = b"STD.P"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    STD_S = b"STD.S"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    TWA = b"TWA"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    VAR_P = b"VAR.P"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    VAR_S = b"VAR.S"

    #: Used by:
    #:
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    BUCKETTIMESTAMP = b"BUCKETTIMESTAMP"

    #: Used by:
    #:
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    EMPTY = b"EMPTY"

    #: Used by:
    #:
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    FILTER_BY_VALUE = b"FILTER_BY_VALUE"

    #: Used by:
    #:
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    GROUPBY = b"GROUPBY"

    #: Used by:
    #:
    #:  - ``TS.MGET``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    SELECTED_LABELS = b"SELECTED_LABELS"

    #: Used by:
    #:
    #:  - ``TS.MGET``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    WITHLABELS = b"WITHLABELS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    GEO = b"GEO"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    HASH = b"HASH"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    JSON = b"JSON"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    MAXTEXTFIELDS = b"MAXTEXTFIELDS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    NOFIELDS = b"NOFIELDS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    NOFREQS = b"NOFREQS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    NOHL = b"NOHL"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    NOINDEX = b"NOINDEX"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    NOOFFSETS = b"NOOFFSETS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    NUMERIC = b"NUMERIC"

    #: Used by:
    #:
    #:  - ``FT.ALTER``
    #:  - ``FT.CREATE``
    SCHEMA = b"SCHEMA"

    #: Used by:
    #:
    #:  - ``FT.ALTER``
    #:  - ``FT.CREATE``
    #:  - ``FT.SYNUPDATE``
    SKIPINITIALSCAN = b"SKIPINITIALSCAN"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    SORTABLE = b"SORTABLE"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    TAG = b"TAG"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    TEXT = b"TEXT"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    UNF = b"UNF"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    VECTOR = b"VECTOR"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    WITHSUFFIXTRIE = b"WITHSUFFIXTRIE"

    #: Used by:
    #:
    #:  - ``FT.ALTER``
    ADD = b"ADD"

    #: Used by:
    #:
    #:  - ``FT.DROPINDEX``
    DELETE_DOCS = b"DD"

    #: Used by:
    #:
    #:  - ``FT.SPELLCHECK``
    EXCLUDE = b"EXCLUDE"

    #: Used by:
    #:
    #:  - ``FT.SPELLCHECK``
    INCLUDE = b"INCLUDE"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    EXPLAINSCORE = b"EXPLAINSCORE"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    HIGHLIGHT = b"HIGHLIGHT"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    INORDER = b"INORDER"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.SEARCH``
    LIMIT = b"LIMIT"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    NOCONTENT = b"NOCONTENT"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    NOSTOPWORDS = b"NOSTOPWORDS"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.SEARCH``
    PARAMS = b"PARAMS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    SUMMARIZE = b"SUMMARIZE"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    TAGS = b"TAGS"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.SEARCH``
    VERBATIM = b"VERBATIM"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``FT.SUGGET``
    WITHPAYLOADS = b"WITHPAYLOADS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    WITHSORTKEYS = b"WITHSORTKEYS"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    LOADALL = b"LOAD *"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    WITHCURSOR = b"WITHCURSOR"

    #: Used by:
    #:
    #:  - ``FT.PROFILE``
    AGGREGATE = b"AGGREGATE"

    #: Used by:
    #:
    #:  - ``FT.PROFILE``
    LIMITED = b"LIMITED"

    #: Used by:
    #:
    #:  - ``FT.PROFILE``
    QUERYWORD = b"QUERY"

    #: Used by:
    #:
    #:  - ``FT.PROFILE``
    SEARCH = b"SEARCH"

    #: Used by:
    #:
    #:  - ``FT.SUGADD``
    INCR = b"INCR"

    #: Used by:
    #:
    #:  - ``FT.SUGGET``
    FUZZY = b"FUZZY"


class PrefixToken(CaseAndEncodingInsensitiveEnum):
    """
    Enum for internal use when adding prefixes to arguments
    """

    #: Used by:
    #:
    #:  - ``CLUSTER SETSLOT``
    #:  - ``GRAPH.CONSTRAINT CREATE``
    #:  - ``GRAPH.CONTRAINT DROP``
    NODE = b"NODE"

    #: Used by:
    #:
    #:  - ``GRAPH.CONSTRAINT CREATE``
    #:  - ``GRAPH.CONTRAINT DROP``
    RELATIONSHIP = b"RELATIONSHIP"

    #: Used by:
    #:
    #:  - ``GRAPH.CONSTRAINT CREATE``
    #:  - ``GRAPH.CONTRAINT DROP``
    PROPERTIES = b"PROPERTIES"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    #:  - ``BITFIELD_RO``
    #:  - ``SORT``
    #:  - ``SORT_RO``
    GET = b"GET"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    INCRBY = b"INCRBY"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    OVERFLOW = b"OVERFLOW"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    SET = b"SET"

    #: Used by:
    #:
    #:  - ``BLMPOP``
    #:  - ``BZMPOP``
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.CURSOR READ``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    #:  - ``HSCAN``
    #:  - ``LMPOP``
    #:  - ``LPOS``
    #:  - ``SCAN``
    #:  - ``SSCAN``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    #:  - ``XAUTOCLAIM``
    #:  - ``XINFO STREAM``
    #:  - ``XRANGE``
    #:  - ``XREAD``
    #:  - ``XREADGROUP``
    #:  - ``XREVRANGE``
    #:  - ``ZMPOP``
    #:  - ``ZSCAN``
    COUNT = b"COUNT"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    ADDR = b"ADDR"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    #:  - ``CLIENT LIST``
    IDENTIFIER = b"ID"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    LADDR = b"LADDR"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    SKIPME = b"SKIPME"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    #:  - ``CLIENT LIST``
    #:  - ``SCAN``
    TYPE = b"TYPE"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    USER = b"USER"

    #: Used by:
    #:
    #:  - ``CLIENT SETINFO``
    LIB_NAME = b"LIB-NAME"

    #: Used by:
    #:
    #:  - ``CLIENT SETINFO``
    LIB_VER = b"LIB-VER"

    #: Used by:
    #:
    #:  - ``CLIENT TRACKING``
    #:  - ``FT.CREATE``
    PREFIX = b"PREFIX"

    #: Used by:
    #:
    #:  - ``CLIENT TRACKING``
    REDIRECT = b"REDIRECT"

    #: Used by:
    #:
    #:  - ``CLUSTER SETSLOT``
    IMPORTING = b"IMPORTING"

    #: Used by:
    #:
    #:  - ``CLUSTER SETSLOT``
    MIGRATING = b"MIGRATING"

    #: Used by:
    #:
    #:  - ``COMMAND LIST``
    ACLCAT = b"ACLCAT"

    #: Used by:
    #:
    #:  - ``COMMAND LIST``
    FILTERBY = b"FILTERBY"

    #: Used by:
    #:
    #:  - ``COMMAND LIST``
    MODULE = b"MODULE"

    #: Used by:
    #:
    #:  - ``COMMAND LIST``
    PATTERN = b"PATTERN"

    #: Used by:
    #:
    #:  - ``COPY``
    DB = b"DB"

    #: Used by:
    #:
    #:  - ``FAILOVER``
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.SEARCH``
    #:  - ``GRAPH.PROFILE``
    #:  - ``GRAPH.QUERY``
    #:  - ``GRAPH.RO_QUERY``
    TIMEOUT = b"TIMEOUT"

    #: Used by:
    #:
    #:  - ``FAILOVER``
    TO = b"TO"

    #: Used by:
    #:
    #:  - ``FUNCTION LIST``
    LIBRARYNAME = b"LIBRARYNAME"

    #: Used by:
    #:
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``SORT``
    STORE = b"STORE"

    #: Used by:
    #:
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    STOREDIST = b"STOREDIST"

    #: Used by:
    #:
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    BYBOX = b"BYBOX"

    #: Used by:
    #:
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    BYRADIUS = b"BYRADIUS"

    #: Used by:
    #:
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    FROMLONLAT = b"FROMLONLAT"

    #: Used by:
    #:
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    FROMMEMBER = b"FROMMEMBER"

    #: Used by:
    #:
    #:  - ``GETEX``
    #:  - ``SET``
    EX = b"EX"

    #: Used by:
    #:
    #:  - ``GETEX``
    #:  - ``SET``
    EXAT = b"EXAT"

    #: Used by:
    #:
    #:  - ``GETEX``
    #:  - ``SET``
    PX = b"PX"

    #: Used by:
    #:
    #:  - ``GETEX``
    #:  - ``SET``
    PXAT = b"PXAT"

    #: Used by:
    #:
    #:  - ``HELLO``
    #:  - ``MIGRATE``
    AUTH = b"AUTH"

    #: Used by:
    #:
    #:  - ``HELLO``
    SETNAME = b"SETNAME"

    #: Used by:
    #:
    #:  - ``HSCAN``
    #:  - ``SCAN``
    #:  - ``SSCAN``
    #:  - ``ZSCAN``
    MATCH = b"MATCH"

    #: Used by:
    #:
    #:  - ``LCS``
    MINMATCHLEN = b"MINMATCHLEN"

    #: Used by:
    #:
    #:  - ``LOLWUT``
    VERSION = b"VERSION"

    #: Used by:
    #:
    #:  - ``LPOS``
    MAXLEN = b"MAXLEN"

    #: Used by:
    #:
    #:  - ``LPOS``
    RANK = b"RANK"

    #: Used by:
    #:
    #:  - ``MEMORY USAGE``
    SAMPLES = b"SAMPLES"

    #: Used by:
    #:
    #:  - ``MIGRATE``
    AUTH2 = b"AUTH2"

    #: Used by:
    #:
    #:  - ``MIGRATE``
    KEYS = b"KEYS"

    #: Used by:
    #:
    #:  - ``MODULE LOADEX``
    ARGS = b"ARGS"

    #: Used by:
    #:
    #:  - ``MODULE LOADEX``
    CONFIG = b"CONFIG"

    #: Used by:
    #:
    #:  - ``RESTORE``
    #:  - ``RESTORE-ASKING``
    FREQ = b"FREQ"

    #: Used by:
    #:
    #:  - ``RESTORE``
    #:  - ``RESTORE-ASKING``
    IDLETIME = b"IDLETIME"

    #: Used by:
    #:
    #:  - ``SINTERCARD``
    #:  - ``SORT``
    #:  - ``SORT_RO``
    #:  - ``XADD``
    #:  - ``XTRIM``
    #:  - ``ZINTERCARD``
    #:  - ``ZRANGE``
    #:  - ``ZRANGEBYLEX``
    #:  - ``ZRANGEBYSCORE``
    #:  - ``ZRANGESTORE``
    #:  - ``ZREVRANGEBYLEX``
    #:  - ``ZREVRANGEBYSCORE``
    LIMIT = b"LIMIT"

    #: Used by:
    #:
    #:  - ``SORT``
    #:  - ``SORT_RO``
    BY = b"BY"

    #: Used by:
    #:
    #:  - ``XCLAIM``
    #:  - ``XPENDING``
    IDLE = b"IDLE"

    #: Used by:
    #:
    #:  - ``XCLAIM``
    LASTID = b"LASTID"

    #: Used by:
    #:
    #:  - ``XCLAIM``
    RETRYCOUNT = b"RETRYCOUNT"

    #: Used by:
    #:
    #:  - ``XCLAIM``
    TIME = b"TIME"

    #: Used by:
    #:
    #:  - ``XGROUP CREATE``
    #:  - ``XGROUP SETID``
    ENTRIESREAD = b"ENTRIESREAD"

    #: Used by:
    #:
    #:  - ``XREAD``
    #:  - ``XREADGROUP``
    BLOCK = b"BLOCK"

    #: Used by:
    #:
    #:  - ``XREAD``
    #:  - ``XREADGROUP``
    STREAMS = b"STREAMS"

    #: Used by:
    #:
    #:  - ``XREADGROUP``
    GROUP = b"GROUP"

    #: Used by:
    #:
    #:  - ``XSETID``
    ENTRIESADDED = b"ENTRIESADDED"

    #: Used by:
    #:
    #:  - ``XSETID``
    MAXDELETEDID = b"MAXDELETEDID"

    #: Used by:
    #:
    #:  - ``ZINTER``
    #:  - ``ZINTERSTORE``
    #:  - ``ZUNION``
    #:  - ``ZUNIONSTORE``
    AGGREGATE = b"AGGREGATE"

    #: Used by:
    #:
    #:  - ``ZINTER``
    #:  - ``ZINTERSTORE``
    #:  - ``ZUNION``
    #:  - ``ZUNIONSTORE``
    WEIGHTS = b"WEIGHTS"

    #: Used by:
    #:
    #:  - ``JSON.GET``
    INDENT = b"INDENT"

    #: Used by:
    #:
    #:  - ``JSON.GET``
    NEWLINE = b"NEWLINE"

    #: Used by:
    #:
    #:  - ``JSON.GET``
    SPACE = b"SPACE"

    #: Used by:
    #:
    #:  - ``BF.INSERT``
    #:  - ``BF.RESERVE``
    #:  - ``CF.RESERVE``
    EXPANSION = b"EXPANSION"

    #: Used by:
    #:
    #:  - ``BF.INSERT``
    #:  - ``CF.INSERT``
    #:  - ``CF.INSERTNX``
    CAPACITY = b"CAPACITY"

    #: Used by:
    #:
    #:  - ``BF.INSERT``
    ERROR = b"ERROR"

    #: Used by:
    #:
    #:  - ``CF.RESERVE``
    BUCKETSIZE = b"BUCKETSIZE"

    #: Used by:
    #:
    #:  - ``CF.RESERVE``
    MAXITERATIONS = b"MAXITERATIONS"

    #: Used by:
    #:
    #:  - ``TDIGEST.CREATE``
    COMPRESSION = b"COMPRESSION"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.DECRBY``
    #:  - ``TS.INCRBY``
    CHUNK_SIZE = b"CHUNK_SIZE"

    #: Used by:
    #:
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    DUPLICATE_POLICY = b"DUPLICATE_POLICY"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.CREATE``
    ENCODING = b"ENCODING"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.DECRBY``
    #:  - ``TS.INCRBY``
    LABELS = b"LABELS"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.DECRBY``
    #:  - ``TS.INCRBY``
    RETENTION = b"RETENTION"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    ON_DUPLICATE = b"ON_DUPLICATE"

    #: Used by:
    #:
    #:  - ``TS.DECRBY``
    #:  - ``TS.INCRBY``
    TIMESTAMP = b"TIMESTAMP"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    AGGREGATION = b"AGGREGATION"

    #: Used by:
    #:
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    ALIGN = b"ALIGN"

    #: Used by:
    #:
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    FILTER_BY_TS = b"FILTER_BY_TS"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.CREATE``
    #:  - ``FT.SEARCH``
    #:  - ``TS.MGET``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    FILTER = b"FILTER"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.CREATE``
    #:  - ``FT.SEARCH``
    AS = b"AS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    #:  - ``FT.SEARCH``
    LANGUAGE = b"LANGUAGE"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    LANGUAGE_FIELD = b"LANGUAGE_FIELD"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    ON = b"ON"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    PAYLOAD_FIELD = b"PAYLOAD_FIELD"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    SCORE = b"SCORE"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    SCORE_FIELD = b"SCORE_FIELD"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    STOPWORDS = b"STOPWORDS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    TEMPORARY = b"TEMPORARY"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.EXPLAIN``
    #:  - ``FT.EXPLAINCLI``
    #:  - ``FT.SEARCH``
    #:  - ``FT.SPELLCHECK``
    DIALECT = b"DIALECT"

    #: Used by:
    #:
    #:  - ``FT.SPELLCHECK``
    DISTANCE = b"DISTANCE"

    #: Used by:
    #:
    #:  - ``FT.SPELLCHECK``
    TERMS = b"TERMS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    EXPANDER = b"EXPANDER"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    FIELDS = b"FIELDS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    FRAGS = b"FRAGS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    GEOFILTER = b"GEOFILTER"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    INFIELDS = b"INFIELDS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    INKEYS = b"INKEYS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    LEN = b"LEN"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``FT.SUGADD``
    PAYLOAD = b"PAYLOAD"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    RETURN = b"RETURN"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    SCORER = b"SCORER"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    SEPARATOR = b"SEPARATOR"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    SLOP = b"SLOP"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.SEARCH``
    SORTBY = b"SORTBY"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    APPLY = b"APPLY"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    GROUPBY = b"GROUPBY"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    LOAD = b"LOAD"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.SUGGET``
    MAX = b"MAX"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    MAXIDLE = b"MAXIDLE"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    REDUCE = b"REDUCE"



from __future__ import annotations

from collections import OrderedDict

from coredis._json import json
from coredis._utils import EncodingInsensitiveDict
from coredis.modules.response.types import (
    SearchAggregationResult,
    SearchDocument,
    SearchResult,
)
from coredis.response._callbacks import ResponseCallback
from coredis.response._utils import flat_pairs_to_dict
from coredis.typing import (
    AnyStr,
    Dict,
    List,
    Optional,
    ResponsePrimitive,
    ResponseType,
    StringT,
    TypedDict,
    ValueT,
)


class SearchConfigCallback(
    ResponseCallback[
        List[List[ResponsePrimitive]],
        List[List[ResponsePrimitive]],
        Dict[AnyStr, ResponsePrimitive],
    ]
):
    def transform(
        self, response: List[List[ResponsePrimitive]], **options: Optional[ValueT]
    ) -> Dict[AnyStr, ResponsePrimitive]:
        pieces = []
        for item in response:
            try:
                v = (item[0], json.loads(item[1]))
            except (ValueError, TypeError):
                v = item
            pieces.append(v)
        return dict(pieces)


class SearchResultCallback(
    ResponseCallback[
        List[ResponseType],
        List[ResponseType],
        SearchResult[AnyStr],
    ]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> SearchResult[AnyStr]:
        step = 2
        results = []
        score_idx = payload_idx = sort_key_idx = 0
        if options.get("withscores"):
            score_idx = 1
            step += 1
        if options.get("withpayloads"):
            payload_idx = score_idx + 1
            step += 1
        if options.get("withsortkeys"):
            sort_key_idx = payload_idx + 1
            step += 1

        for k in range(1, len(response) - 1, step):
            section = response[k : k + step]
            score_explain = None
            if options.get("explainscore"):
                score = section[score_idx][0]
                score_explain = section[score_idx][1]
            else:
                score = section[score_idx] if score_idx else None
            fields = EncodingInsensitiveDict(flat_pairs_to_dict(section[-1]))
            if "$" in fields:
                fields = json.loads(fields.pop("$"))
            results.append(
                SearchDocument(
                    section[0],
                    float(score) if score else None,
                    score_explain,
                    section[payload_idx] if payload_idx else None,
                    section[sort_key_idx] if sort_key_idx else None,
                    fields,
                )
            )
        return SearchResult[AnyStr](response[0], tuple(results))


class AggregationResultCallback(
    ResponseCallback[
        List[ResponseType],
        List[ResponseType],
        SearchAggregationResult[AnyStr],
    ]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> SearchAggregationResult:
        def try_json(value):
            if not options.get("dialect", None) == 3:
                return value
            try:
                return json.loads(value)
            except ValueError:
                return value

        return SearchAggregationResult[AnyStr](
            [
                flat_pairs_to_dict(k, try_json)
                for k in (
                    response[1:] if not options.get("with_cursor") else response[0][1:]
                )
            ],
            response[1] if options.get("with_cursor") else None,
        )


class SpellCheckResult(TypedDict):
    term: StringT
    suggestions: OrderedDict[StringT, int]


class SpellCheckCallback(
    ResponseCallback[
        List[ResponseType],
        List[ResponseType],
        SpellCheckResult,
    ]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> SpellCheckResult:
        suggestions = {}
        for result in response:
            suggestions[result[1]] = OrderedDict((k[1], float(k[0])) for k in result[2])

        return suggestions



from __future__ import annotations

from typing import Any

from coredis._utils import EncodingInsensitiveDict
from coredis.response._callbacks import ResponseCallback
from coredis.response._utils import flat_pairs_to_dict, flat_pairs_to_ordered_dict
from coredis.response.types import (
    StreamEntry,
    StreamInfo,
    StreamPending,
    StreamPendingExt,
)
from coredis.typing import (
    AnyStr,
    Dict,
    Optional,
    OrderedDict,
    ResponseType,
    StringT,
    Tuple,
    Union,
    ValueT,
)


class StreamRangeCallback(
    ResponseCallback[ResponseType, ResponseType, Tuple[StreamEntry, ...]]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Tuple[StreamEntry, ...]:
        return tuple(
            StreamEntry(r[0], flat_pairs_to_ordered_dict(r[1])) for r in response
        )


class ClaimCallback(
    ResponseCallback[
        ResponseType, ResponseType, Union[Tuple[AnyStr, ...], Tuple[StreamEntry, ...]]
    ]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Union[Tuple[AnyStr, ...], Tuple[StreamEntry, ...]]:
        if options.get("justid") is not None:
            return tuple(response)
        else:
            return StreamRangeCallback()(response)


class AutoClaimCallback(
    ResponseCallback[
        ResponseType,
        ResponseType,
        Union[
            Tuple[AnyStr, Tuple[AnyStr, ...]],
            Tuple[AnyStr, Tuple[StreamEntry, ...], Tuple[AnyStr, ...]],
        ],
    ]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Union[
        Tuple[AnyStr, Tuple[AnyStr, ...]],
        Tuple[AnyStr, Tuple[StreamEntry, ...], Tuple[AnyStr, ...]],
    ]:
        if options.get("justid") is not None:
            return response[0], tuple(response[1])
        else:
            return (
                response[0],
                StreamRangeCallback()(response[1]),
                tuple(response[2]) if len(response) > 2 else (),
            )


class MultiStreamRangeCallback(
    ResponseCallback[
        ResponseType, ResponseType, Optional[Dict[AnyStr, Tuple[StreamEntry, ...]]]
    ]
):
    def transform_3(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Optional[Dict[AnyStr, Tuple[StreamEntry, ...]]]:
        if response:
            mapping: Dict[AnyStr, Tuple[StreamEntry, ...]] = {}

            for stream_id, entries in response.items():
                mapping[stream_id] = tuple(
                    StreamEntry(r[0], flat_pairs_to_ordered_dict(r[1])) for r in entries
                )

            return mapping
        return None

    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Optional[Dict[AnyStr, Tuple[StreamEntry, ...]]]:
        if response:
            mapping: Dict[AnyStr, Tuple[StreamEntry, ...]] = {}

            for stream_id, entries in response:
                mapping[stream_id] = tuple(
                    StreamEntry(r[0], flat_pairs_to_ordered_dict(r[1])) for r in entries
                )

            return mapping
        return None


class PendingCallback(
    ResponseCallback[
        ResponseType, ResponseType, Union[StreamPending, Tuple[StreamPendingExt, ...]]
    ]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Union[StreamPending, Tuple[StreamPendingExt, ...]]:
        if not options.get("count"):
            return StreamPending(
                response[0],
                response[1],
                response[2],
                OrderedDict((r[0], int(r[1])) for r in response[3] or []),
            )
        else:
            return tuple(
                StreamPendingExt(sub[0], sub[1], sub[2], sub[3]) for sub in response
            )


class XInfoCallback(
    ResponseCallback[ResponseType, ResponseType, Tuple[Dict[AnyStr, AnyStr], ...]]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Tuple[Dict[AnyStr, AnyStr], ...]:
        return tuple(flat_pairs_to_dict(row) for row in response)


class StreamInfoCallback(ResponseCallback[ResponseType, ResponseType, StreamInfo]):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> StreamInfo:
        res: Dict[StringT, Any] = EncodingInsensitiveDict(flat_pairs_to_dict(response))
        if not options.get("full"):
            k1 = "first-entry"
            kn = "last-entry"
            e1: Optional[StreamEntry] = None
            en: Optional[StreamEntry] = None

            if len(res.get(k1, [])) > 0:
                v = res.get(k1)
                e1 = StreamEntry(v[0], flat_pairs_to_ordered_dict(v[1]))
                res.pop(k1)

            if len(res.get(kn, [])) > 0:
                v = res.get(kn)
                en = StreamEntry(v[0], flat_pairs_to_ordered_dict(v[1]))
                res.pop(kn)
            res.update({"first-entry": e1, "last-entry": en})
        else:
            groups = res.get("groups")
            if groups:
                normalized_groups = []
                for group in groups:
                    g = EncodingInsensitiveDict(flat_pairs_to_dict(group))
                    consumers = g["consumers"]
                    normalized_consumers = []
                    for consumer in consumers:
                        normalized_consumers.append(flat_pairs_to_dict(consumer))
                    g["consumers"] = normalized_consumers
                    normalized_groups.append(g)
                res["groups"] = normalized_groups
            res.update(
                {
                    "entries": tuple(
                        StreamEntry(k[0], flat_pairs_to_ordered_dict(k[1]))
                        for k in res.get("entries", [])
                    )
                }
            )
        stream_info: StreamInfo = {
            "first-entry": res.get("first-entry"),
            "last-entry": res.get("last-entry"),
            "length": res["length"],
            "radix-tree-keys": res["radix-tree-keys"],
            "radix-tree-nodes": res["radix-tree-nodes"],
            "groups": res["groups"],
            "last-generated-id": res["last-generated-id"],
            "max-deleted-entry-id": str(res.get("max-deleted-entry-id")),
            "entries-added": int(res.get("entries-added", 0)),
            "recorded-first-entry-id": str(res.get("recorded-first-entry-id")),
            "entries-read": int(res.get("entries-read", 0)),
            "entries": res.get("entries"),
        }
        return stream_info



from __future__ import annotations

import contextlib
import dataclasses
import functools
import random
import textwrap
import warnings
from typing import TYPE_CHECKING, Any, cast

from packaging import version

from coredis.cache import AbstractCache, SupportsSampling
from coredis.commands._utils import check_version, redis_command_link
from coredis.commands.constants import CommandFlag, CommandGroup, CommandName, NodeFlag
from coredis.globals import COMMAND_FLAGS, READONLY_COMMANDS
from coredis.response._callbacks import ClusterMultiNodeCallback
from coredis.typing import (
    AsyncIterator,
    Callable,
    Coroutine,
    Dict,
    NamedTuple,
    Optional,
    P,
    R,
    ResponseType,
    Set,
    add_runtime_checks,
)

if TYPE_CHECKING:
    pass


@dataclasses.dataclass
class CacheConfig:
    key_func: Callable[..., bytes]


class RedirectUsage(NamedTuple):
    reason: str
    warn: bool


@dataclasses.dataclass
class CommandDetails:
    command: bytes
    group: Optional[CommandGroup]
    version_introduced: Optional[version.Version]
    version_deprecated: Optional[version.Version]
    _arguments: Optional[Dict[str, Dict[str, str]]]
    cluster: ClusterCommandConfig
    cache_config: Optional[CacheConfig]
    flags: Set[CommandFlag]
    redirect_usage: Optional[RedirectUsage]
    arguments: Dict[str, version.Version] = dataclasses.field(
        init=False, default_factory=lambda: {}
    )

    def __post_init__(self) -> None:
        self.arguments = {
            k: version.Version(v["version_introduced"])
            for k, v in (self._arguments or {}).items()
            if v.get("version_introduced")
        }


@dataclasses.dataclass
class ClusterCommandConfig:
    enabled: bool = True
    combine: Optional[ClusterMultiNodeCallback] = None  # type: ignore
    route: Optional[NodeFlag] = None
    split: Optional[NodeFlag] = None

    @property
    def multi_node(self) -> bool:
        return (self.route or self.split) in [
            NodeFlag.ALL,
            NodeFlag.PRIMARIES,
            NodeFlag.REPLICAS,
        ]


@dataclasses.dataclass
class CommandCache:
    command: bytes
    cache_config: Optional[CacheConfig]

    @contextlib.asynccontextmanager
    async def __call__(
        self,
        func: Callable[P, Coroutine[Any, Any, R]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> AsyncIterator[R]:
        from coredis.modules.base import ModuleGroup

        client = args[0]
        if isinstance(args[0], ModuleGroup):
            client = args[0].client

        cache = getattr(client, "cache")
        noreply = getattr(client, "noreply")
        if not (self.cache_config and cache) or noreply:
            yield await func(*args, **kwargs)
        else:
            assert isinstance(cache, AbstractCache)
            if not cache.healthy:
                yield await func(*args, **kwargs)
            else:
                key = self.cache_config.key_func(*args[1:], **kwargs)
                try:
                    cached = cast(
                        R,
                        cache.get(
                            self.command, key, *args[1:], *kwargs.items()  # type: ignore
                        ),
                    )
                    if isinstance(
                        cache, SupportsSampling
                    ) and not random.random() * 100.0 < min(100.0, cache.confidence):
                        actual = await func(*args, **kwargs)
                        cache.feedback(
                            self.command,
                            key,
                            *args[1:],  # type: ignore
                            *kwargs.items(),  # type: ignore
                            match=(actual == cached),
                        )
                        yield actual
                    else:
                        yield cached
                except KeyError:
                    response = await func(*args, **kwargs)
                    cache.put(
                        self.command,
                        key,
                        *args[1:],  # type: ignore
                        *kwargs.items(),  # type: ignore
                        value=cast(ResponseType, response),
                    )
                    yield response


def redis_command(
    command_name: CommandName,
    group: Optional[CommandGroup] = None,
    version_introduced: Optional[str] = None,
    version_deprecated: Optional[str] = None,
    deprecation_reason: Optional[str] = None,
    redirect_usage: Optional[RedirectUsage] = None,
    arguments: Optional[Dict[str, Dict[str, str]]] = None,
    flags: Optional[Set[CommandFlag]] = None,
    cluster: ClusterCommandConfig = ClusterCommandConfig(),
    cache_config: Optional[CacheConfig] = None,
) -> Callable[
    [Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]
]:
    readonly = False
    if flags and CommandFlag.READONLY in flags:
        READONLY_COMMANDS.add(command_name)
        readonly = True

    if not readonly and cache_config:  # noqa
        raise RuntimeError(
            f"Can't decorate non readonly command {command_name} with cache config"
        )

    COMMAND_FLAGS[command_name] = flags or set()

    command_details = CommandDetails(
        command_name,
        group,
        version.Version(version_introduced) if version_introduced else None,
        version.Version(version_deprecated) if version_deprecated else None,
        arguments,
        cluster or ClusterCommandConfig(),
        cache_config,
        flags or set(),
        redirect_usage,
    )

    def wrapper(
        func: Callable[P, Coroutine[Any, Any, R]]
    ) -> Callable[P, Coroutine[Any, Any, R]]:
        command_cache = CommandCache(command_name, cache_config)
        runtime_checkable = add_runtime_checks(func)

        @functools.wraps(func)
        async def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            from coredis.client import Redis, RedisCluster

            client = args[0]
            is_regular_client = isinstance(client, (Redis, RedisCluster))
            runtime_checking = (
                not getattr(client, "noreply", None) and is_regular_client
            )
            if redirect_usage and is_regular_client:
                if redirect_usage.warn:
                    warnings.warn(redirect_usage.reason, UserWarning, stacklevel=2)
                else:
                    raise NotImplementedError(redirect_usage.reason)
            callable = runtime_checkable if runtime_checking else func
            await check_version(
                client,  # type: ignore
                func.__name__,
                command_details,
                deprecation_reason,
                kwargs,
            )
            async with command_cache(callable, *args, **kwargs) as response:
                return response

        wrapped.__doc__ = textwrap.dedent(wrapped.__doc__ or "")
        if group:
            wrapped.__doc__ = f"""
{wrapped.__doc__}

Redis command documentation: {redis_command_link(command_name)}
"""
        if version_introduced or command_details.arguments:
            wrapped.__doc__ += """
Compatibility:
"""

        if version_introduced:
            wrapped.__doc__ += f"""
- New in :redis-version:`{version_introduced}`
"""
        if version_deprecated and deprecation_reason:
            wrapped.__doc__ += f"""
- Deprecated in :redis-version:`{version_deprecated}`
      {deprecation_reason.strip()}
"""
        elif version_deprecated:
            wrapped.__doc__ += f"""
- Deprecated in :redis-version:`{version_deprecated}`
"""
        if command_details.arguments:
            for argument, min_version in command_details.arguments.items():
                wrapped.__doc__ += f"""
- :paramref:`{argument}`: New in :redis-version:`{min_version}`
"""
        if cache_config:
            wrapped.__doc__ += """
.. hint:: Supports client side caching
"""
        if redirect_usage:
            if redirect_usage.warn:
                preamble = f".. warning:: Using ``{func.__name__}`` directly is not recommended."
            else:
                preamble = (
                    f".. danger:: Using ``{func.__name__}`` directly is not supported."
                )
            wrapped.__doc__ += f"""
{preamble} {redirect_usage.reason}
"""

        setattr(wrapped, "__coredis_command", command_details)
        return wrapped

    return wrapper



from __future__ import annotations

import os
import pathlib
import platform
import sys

import versioneer

__author__ = "Ali-Akber Saifee"
__email__ = "ali@indydevs.org"
__copyright__ = "Copyright 2023, Ali-Akber Saifee"

from setuptools import find_packages, setup
from setuptools.command.build_ext import build_ext
from setuptools.extension import Extension

THIS_DIR = os.path.abspath(os.path.dirname(__file__))
PY_IMPLEMENTATION = platform.python_implementation()
USE_MYPYC = False
PURE_PYTHON = os.environ.get("PURE_PYTHON", PY_IMPLEMENTATION != "CPython")


def get_requirements(req_file):
    requirements = []

    for r in open(os.path.join(THIS_DIR, "requirements", req_file)).read().splitlines():
        req = r.strip()

        if req.startswith("-r"):
            requirements.extend(get_requirements(req.replace("-r ", "")))
        elif req:
            requirements.append(req)

    return requirements


class coredis_build_ext(build_ext):
    warning_message = """
********************************************************************
{target} could not
be compiled. No C extensions are essential for coredis to run,
although they do result in significant speed improvements for
response parsing.
{comment}
********************************************************************
"""

    def run(self):
        try:
            super().run()
        except Exception as e:
            self.warn(e)
            self.warn(
                self.warning_message.format(
                    target="Extension modules",
                    comment=(
                        "There is an issue with your platform configuration "
                        "- see above."
                    ),
                )
            )

    def build_extension(self, ext):
        try:
            super().build_extension(ext)
        except Exception as e:
            self.warn(e)
            self.warn(
                self.warning_message.format(
                    target=f"The {ext.name} extension ",
                    comment=(
                        "The output above this warning shows how the "
                        "compilation failed."
                    ),
                )
            )


_ROOT_DIR = pathlib.Path(__file__).parent

with open(str(_ROOT_DIR / "README.md")) as f:
    long_description = f.read()

if len(sys.argv) > 1 and "--use-mypyc" in sys.argv:
    sys.argv.remove("--use-mypyc")
    USE_MYPYC = True

if not PURE_PYTHON:
    extensions = [
        Extension(
            name="coredis.speedups",
            sources=["coredis/speedups.c"],
        )
    ]

    if USE_MYPYC:
        from mypyc.build import mypycify

        extensions += mypycify(
            [
                "coredis/constants.py",
                "coredis/parser.py",
                "coredis/_packer.py",
            ],
            debug_level="0",
            strip_asserts=True,
        )
        for ext in extensions:
            if "-Werror" in ext.extra_compile_args:
                ext.extra_compile_args.remove("-Werror")
else:
    extensions = []


setup(
    name="coredis",
    version=versioneer.get_version(),
    description="Python async client for Redis key-value store",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alisaifee/coredis",
    project_urls={
        "Source": "https://github.com/alisaifee/coredis",
        "Changes": "https://github.com/alisaifee/coredis/releases",
        "Documentation": "https://coredis.readthedocs.org",
    },
    author=__author__,
    author_email=__email__,
    maintainer=__author__,
    maintainer_email=__email__,
    keywords=["Redis", "key-value store", "asyncio"],
    license="MIT",
    packages=find_packages(exclude=["*tests*"]),
    include_package_data=True,
    package_data={
        "coredis": ["py.typed"],
    },
    python_requires=">=3.7",
    install_requires=get_requirements("main.txt"),
    cmdclass=versioneer.get_cmdclass(
        {
            "build_ext": coredis_build_ext,
        }
    ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    ext_modules=extensions,
)



from __future__ import annotations

import asyncio
import contextvars
import importlib.resources
import math
import time
import uuid
import warnings
from types import TracebackType
from typing import cast

from coredis.client import Redis, RedisCluster
from coredis.commands import Script
from coredis.exceptions import LockError, ReplicationError
from coredis.tokens import PureToken
from coredis.typing import AnyStr, Generic, KeyT, Optional, StringT, Type, Union

with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)
    EXTEND_SCRIPT = Script(
        script=importlib.resources.read_text(__package__, "extend.lua")
    )
    RELEASE_SCRIPT = Script(
        script=importlib.resources.read_text(__package__, "release.lua")
    )


class LuaLock(Generic[AnyStr]):
    """
    A shared, distributed Lock using LUA scripts.

    The lock can be used with both :class:`coredis.Redis`
    and :class:`coredis.RedisCluster` either explicitly or as an async context
    manager::


        import asyncio
        import coredis
        from coredis.exceptions import LockError
        from coredis.recipes.locks import LuaLock

        async def test():
            client = coredis.Redis()
            async with LuaLock(client, "mylock", timeout=1.0):
                # do stuff
                await asyncio.sleep(0.5)
                # lock is implictly released when the context manager exits
            try:
                async with LuaLock(client, "mylock", timeout=1.0):
                    # do stuff that takes too long
                    await asyncio.sleep(1)
                    # lock will raise upon exiting the context manager
            except LockError as err:
                # roll back stuff
                print(f"Expected error: {err}")
            lock = LuaLock(client, "mylock", timeout=1.0)
            await lock.acquire()
            # do stuff
            await asyncio.sleep(0.5)
            # do more stuff
            await lock.extend(1.0)
            await lock.release()

        asyncio.run(test())
    """

    @classmethod
    @RELEASE_SCRIPT.wraps(client_arg="client")
    async def lua_release(  # type: ignore[empty-body]
        cls,
        client: Union[Redis[AnyStr], RedisCluster[AnyStr]],
        name: KeyT,
        expected_token: StringT,
    ) -> int:
        ...

    @classmethod
    @EXTEND_SCRIPT.wraps(client_arg="client")
    async def lua_extend(  # type: ignore[empty-body]
        cls,
        client: Union[Redis[AnyStr], RedisCluster[AnyStr]],
        name: KeyT,
        expected_token: StringT,
        additional_time: int,
    ) -> int:
        ...

    local: contextvars.ContextVar[Optional[StringT]]

    def __init__(
        self,
        client: Union[Redis[AnyStr], RedisCluster[AnyStr]],
        name: StringT,
        timeout: Optional[float] = None,
        sleep: float = 0.1,
        blocking: bool = True,
        blocking_timeout: Optional[float] = None,
    ):
        """
        :param timeout: indicates a maximum life for the lock.
         By default, it will remain locked until :meth:`release` is called.
         ``timeout`` can be specified as a float or integer, both representing
         the number of seconds to wait.

        :param sleep: indicates the amount of time to sleep per loop iteration
         when the lock is in blocking mode and another client is currently
         holding the lock.

        :param blocking: indicates whether calling :meth:`acquire` should block until
         the lock has been acquired or to fail immediately, causing :meth:`acquire`
         to return ``False`` and the lock not being acquired. Defaults to ``True``.

        :param blocking_timeout: indicates the maximum amount of time in seconds to
         spend trying to acquire the lock. A value of ``None`` indicates
         continue trying forever. ``blocking_timeout`` can be specified as a
         :class:`float` or :class:`int`, both representing the number of seconds to wait.
        """
        self.client: Union[Redis[AnyStr], RedisCluster[AnyStr]] = client
        self.name = name
        self.timeout = timeout
        self.sleep = sleep
        self.blocking = blocking
        self.blocking_timeout = blocking_timeout
        self.local = contextvars.ContextVar[Optional[StringT]]("token", default=None)
        if self.timeout and self.sleep > self.timeout:
            raise LockError("'sleep' must be less than 'timeout'")

    async def __aenter__(
        self,
    ) -> LuaLock[AnyStr]:
        if await self.acquire():
            return self
        raise LockError("Could not acquire lock")

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.release()

    async def acquire(
        self,
    ) -> bool:
        """
        Use :rediscommand:`SET` with the ``NX`` option
        to acquire a lock. If the lock is being used with a cluster client
        the :meth:`coredis.RedisCluster.ensure_replication` context manager
        will be used to ensure that the command was replicated to atleast
        half the replicas of the shard where the lock would be acquired.

        :raises: :exc:`~coredis.exceptions.LockError`
        """
        token = uuid.uuid1().hex
        blocking = self.blocking
        blocking_timeout = self.blocking_timeout
        stop_trying_at = None
        if blocking_timeout is not None:
            stop_trying_at = time.time() + blocking_timeout
        while True:
            if await self.__acquire(token, stop_trying_at):
                self.local.set(token)
                return True
            if not blocking:
                return False
            if stop_trying_at is not None and time.time() > stop_trying_at:
                return False
            await asyncio.sleep(self.sleep)

    async def release(self) -> None:
        """
        Releases the already acquired lock

        :raises: :exc:`~coredis.exceptions.LockError`
        """
        expected_token = self.local.get()
        if expected_token is None:
            raise LockError("Cannot release an unlocked lock")
        self.local.set(None)
        await self.__release(expected_token)

    async def extend(self, additional_time: float) -> bool:
        """
        Adds more time to an already acquired lock.

        :param additional_time: can be specified as an integer or a float, both
         representing the number of seconds to add.

        :raises: :exc:`~coredis.exceptions.LockError`
        """
        if self.local.get() is None:
            raise LockError("Cannot extend an unlocked lock")
        if self.timeout is None:
            raise LockError("Cannot extend a lock with no timeout")
        return await self.__extend(additional_time)

    @property
    def replication_factor(self) -> int:
        """
        Number of replicas the lock needs to replicate to, to be
        considered acquired.
        """
        if isinstance(self.client, RedisCluster):
            return math.ceil(self.client.num_replicas_per_shard / 2)
        return 0

    async def __acquire(self, token: StringT, stop_trying_at: Optional[float]) -> bool:
        if isinstance(self.client, RedisCluster):
            try:
                replication_wait = (
                    1000 * (max(0, stop_trying_at - time.time()))
                    if stop_trying_at is not None
                    else 100
                )
                with self.client.ensure_replication(
                    self.replication_factor, timeout_ms=int(replication_wait)
                ):
                    return await self.client.set(
                        self.name,
                        token,
                        condition=PureToken.NX,
                        px=int(self.timeout * 1000) if self.timeout else None,
                    )
            except ReplicationError:
                warnings.warn(
                    f"Unable to ensure lock {self.name!r} was replicated "
                    f"to {self.replication_factor} replicas",
                    category=RuntimeWarning,
                )
                await self.client.delete([self.name])
                return False
        else:
            return await self.client.set(
                self.name,
                token,
                condition=PureToken.NX,
                px=int(self.timeout * 1000) if self.timeout else None,
            )

    async def __release(self, expected_token: StringT) -> None:
        if not bool(
            await self.lua_release(
                self.client,
                self.name,
                expected_token,
            )
        ):
            raise LockError("Cannot release a lock that's no longer owned")

    async def __extend(self, additional_time: float) -> bool:
        additional_time = int(additional_time * 1000)
        if additional_time < 0:
            return True
        if not bool(
            await self.lua_extend(
                self.client,
                self.name,
                cast(AnyStr, self.local.get()),
                additional_time,
            )
        ):
            raise LockError("Cannot extend a lock that's no longer owned")
        return True



from __future__ import annotations

import asyncio
import os
import random
import threading
import time
import warnings
from typing import Any, List, cast

import async_timeout

from coredis._utils import b, hash_slot
from coredis.connection import ClusterConnection, Connection
from coredis.exceptions import ConnectionError, RedisClusterException
from coredis.globals import READONLY_COMMANDS
from coredis.pool.basic import ConnectionPool
from coredis.pool.nodemanager import ManagedNode, NodeManager
from coredis.typing import (
    Callable,
    ClassVar,
    Dict,
    Iterable,
    Node,
    Optional,
    Set,
    StringT,
    Type,
    Union,
    ValueT,
)


class ClusterConnectionPool(ConnectionPool):
    """
    Custom connection pool for :class:`~coredis.RedisCluster` client
    """

    #: Mapping of querystring arguments to their parser functions
    URL_QUERY_ARGUMENT_PARSERS: ClassVar[
        Dict[str, Callable[..., Optional[Union[int, float, bool, str]]]]
    ] = {
        **ConnectionPool.URL_QUERY_ARGUMENT_PARSERS,
        "max_connections_per_node": bool,
        "reinitialize_steps": int,
        "skip_full_coverage_check": bool,
        "read_from_replicas": bool,
        "blocking": bool,
    }

    nodes: NodeManager
    connection_class: Type[ClusterConnection]

    _created_connections_per_node: Dict[str, int]
    _cluster_available_connections: Dict[str, asyncio.Queue[Optional[Connection]]]
    _cluster_in_use_connections: Dict[str, Set[Connection]]

    def __init__(
        self,
        startup_nodes: Optional[Iterable[Node]] = None,
        connection_class: Type[ClusterConnection] = ClusterConnection,
        queue_class: Type[asyncio.Queue[Optional[Connection]]] = asyncio.LifoQueue,
        max_connections: Optional[int] = None,
        max_connections_per_node: bool = False,
        reinitialize_steps: Optional[int] = None,
        skip_full_coverage_check: bool = False,
        nodemanager_follow_cluster: bool = True,
        readonly: bool = False,
        read_from_replicas: bool = False,
        max_idle_time: int = 0,
        idle_check_interval: int = 1,
        blocking: bool = False,
        timeout: int = 20,
        **connection_kwargs: Optional[Any],
    ):
        """

        Changes
          - .. versionchanged:: 4.4.0

            - :paramref:`nodemanager_follow_cluster` now defaults to ``True``

          - .. deprecated:: 4.4.0

            - :paramref:`readonly` renamed to :paramref:`read_from_replicas`

        :param max_connections: Maximum number of connections to allow concurrently from this
         client. If the value is ``None`` it will default to 32.
        :param max_connections_per_node: Whether to use the value of :paramref:`max_connections`
         on a per node basis or cluster wide. If ``False`` and :paramref:`blocking` is ``True``
         the per-node connection pools will have a maximum size of :paramref:`max_connections`
         divided by the number of nodes in the cluster.
        :param blocking: If ``True`` the client will block at most :paramref:`timeout` seconds
         if :paramref:`max_connections` is reachd when trying to obtain a connection
        :param timeout: Number of seconds to block if :paramref:`block` is ``True`` when trying to
         obtain a connection.
        :param skip_full_coverage_check:
            Skips the check of cluster-require-full-coverage config, useful for clusters
            without the :rediscommand:`CONFIG` command (For example with AWS Elasticache)
        :param nodemanager_follow_cluster:
            The node manager will during initialization try the last set of nodes that
            it was operating on. This will allow the client to drift along side the cluster
            if the cluster nodes move around alot.
        :param read_from_replicas: If ``True`` the client will route readonly commands to replicas
        """
        super().__init__(
            connection_class=connection_class, max_connections=max_connections
        )
        self.queue_class = queue_class
        # Special case to make from_url method compliant with cluster setting.
        # from_url method will send in the ip and port through a different variable then the
        # regular startup_nodes variable.

        if startup_nodes is None:
            host = connection_kwargs.pop("host", None)
            port = connection_kwargs.pop("port", None)
            if host and port:
                startup_nodes = [Node(host=str(host), port=int(port))]
        self.blocking = blocking
        self.blocking_timeout = timeout
        self.max_connections = max_connections or 2**31
        self.max_connections_per_node = max_connections_per_node
        self.nodes = NodeManager(
            startup_nodes,
            reinitialize_steps=reinitialize_steps,
            skip_full_coverage_check=skip_full_coverage_check,
            max_connections=self.max_connections,
            nodemanager_follow_cluster=nodemanager_follow_cluster,
            **connection_kwargs,  # type: ignore
        )
        self.connection_kwargs = connection_kwargs
        self.connection_kwargs["read_from_replicas"] = read_from_replicas
        self.read_from_replicas = read_from_replicas or readonly
        self.max_idle_time = max_idle_time
        self.idle_check_interval = idle_check_interval
        self.reset()

        if "stream_timeout" not in self.connection_kwargs:
            self.connection_kwargs["stream_timeout"] = None
        self._init_lock = asyncio.Lock()

    def __repr__(self) -> str:
        """
        Returns a string with all unique ip:port combinations that this pool
        is connected to
        """

        return "{}<{}>".format(
            type(self).__name__,
            ", ".join(
                [
                    self.connection_class.describe(node.__dict__)
                    for node in self.nodes.startup_nodes
                ]
            ),
        )

    async def initialize(self) -> None:
        if not self.initialized:
            async with self._init_lock:
                if not self.initialized:
                    await self.nodes.initialize()
                    if (
                        not self.max_connections_per_node
                        and self.max_connections < len(self.nodes.nodes)
                    ):
                        warnings.warn(
                            f"The value of max_connections={self.max_connections} "
                            "should be atleast equal to the number of nodes "
                            f"({len(self.nodes.nodes)}) in the cluster and has been increased by "
                            f"{len(self.nodes.nodes)-self.max_connections} connections."
                        )
                        self.max_connections = len(self.nodes.nodes)
                    await super().initialize()

    async def disconnect_on_idle_time_exceeded(self, connection: Connection) -> None:
        assert isinstance(connection, ClusterConnection)
        while True:
            if (
                time.time() - connection.last_active_at > self.max_idle_time
                and not connection.requests_pending
            ):
                connection.disconnect()
                node = connection.node
                if node.name in self._created_connections_per_node:
                    self._created_connections_per_node[node.name] -= 1
                break
            await asyncio.sleep(self.idle_check_interval)

    def reset(self) -> None:
        """Resets the connection pool back to a clean state"""
        self.pid = os.getpid()
        self._created_connections_per_node = {}
        self._cluster_available_connections = {}
        self._cluster_in_use_connections = {}
        self._check_lock = threading.Lock()
        self.initialized = False

    def checkpid(self) -> None:  # noqa
        if self.pid != os.getpid():
            with self._check_lock:
                if self.pid == os.getpid():
                    # another thread already did the work while we waited
                    # on the lockself.

                    return
                self.disconnect()
                self.reset()

    async def get_connection(
        self,
        command_name: Optional[bytes] = None,
        *keys: ValueT,
        acquire: bool = True,
        **options: Optional[ValueT],
    ) -> Connection:
        # Only pubsub command/connection should be allowed here

        if command_name != b"pubsub":
            raise RedisClusterException(
                "Only 'pubsub' commands can use get_connection()"
            )

        routing_key = options.pop("channel", None)
        node_type = options.pop("node_type", "primary")

        if not routing_key:
            return await self.get_random_connection()

        slot = hash_slot(b(routing_key))
        if node_type == "replica":
            node = self.get_replica_node_by_slot(slot)
        else:
            node = self.get_primary_node_by_slot(slot)
        self.checkpid()

        try:
            connection = self.__node_pool(node.name).get_nowait()
        except asyncio.QueueEmpty:
            connection = None
        if not connection:
            connection = self._make_node_connection(node)
        else:
            if connection.needs_handshake:
                await connection.perform_handshake()

        if acquire:
            self._cluster_in_use_connections.setdefault(node.name, set())
            self._cluster_in_use_connections[node.name].add(connection)
        else:
            self.__node_pool(node.name).put_nowait(connection)
        return connection

    def _make_node_connection(self, node: ManagedNode) -> Connection:
        """Creates a new connection to a node"""

        if self.count_all_num_connections(node) >= self.max_connections:
            if self.max_connections_per_node:
                raise ConnectionError(
                    "Too many connection ({}) for node: {}".format(
                        self.count_all_num_connections(node), node.name
                    )
                )

            raise ConnectionError("Too many connections")

        self._created_connections_per_node.setdefault(node.name, 0)
        self._created_connections_per_node[node.name] += 1
        connection = self.connection_class(
            host=node.host,
            port=node.port,
            **self.connection_kwargs,  # type: ignore
        )

        # Must store node in the connection to make it eaiser to track
        connection.node = node

        if self.max_idle_time > self.idle_check_interval > 0:
            # do not await the future
            asyncio.ensure_future(self.disconnect_on_idle_time_exceeded(connection))

        return connection

    def __node_pool(self, node: str) -> asyncio.Queue[Optional[Connection]]:
        if not self._cluster_available_connections.get(node):
            self._cluster_available_connections[node] = self.__default_node_queue()
        return self._cluster_available_connections[node]

    def __default_node_queue(
        self,
    ) -> asyncio.Queue[Optional[Connection]]:
        q_size = max(
            1,
            int(
                self.max_connections
                if self.max_connections_per_node
                else self.max_connections / len(self.nodes.nodes)
            ),
        )

        q: asyncio.Queue[Optional[Connection]] = self.queue_class(q_size)

        # If the queue is non-blocking, we don't need to pre-populate it
        if not self.blocking:
            return q

        if q_size > 2**16:  # noqa
            raise RuntimeError(
                f"Requested unsupported value of max_connections: {q_size} in blocking mode"
            )

        while True:
            try:
                q.put_nowait(None)
            except asyncio.QueueFull:
                break
        return q

    def release(self, connection: Connection) -> None:
        """Releases the connection back to the pool"""
        assert isinstance(connection, ClusterConnection)

        self.checkpid()

        if connection.pid == self.pid:
            # Remove the current connection from _in_use_connection and add it back to the available
            # pool. There is cases where the connection is to be removed but it will not exist and
            # there must be a safe way to remove
            i_c = self._cluster_in_use_connections.get(connection.node.name, set())

            if connection in i_c:
                i_c.remove(connection)
            else:
                pass
            try:
                self.__node_pool(connection.node.name).put_nowait(connection)
            except asyncio.QueueFull:
                connection.disconnect()
                # reduce node connection count in case of too many connection error raised
                if connection.node.name in self._created_connections_per_node:
                    self._created_connections_per_node[connection.node.name] -= 1

    def disconnect(self) -> None:
        """Closes all connections in the pool"""
        for node_connections in self._cluster_in_use_connections.values():
            for connection in node_connections:
                connection.disconnect()
        for node, available_connections in self._cluster_available_connections.items():
            removed = 0
            while True:
                try:
                    _connection = available_connections.get_nowait()
                    if _connection:
                        _connection.disconnect()
                        if node in self._created_connections_per_node:
                            self._created_connections_per_node[node] -= 1
                    removed += 1
                except asyncio.QueueEmpty:
                    break
            # Refill queue with empty slots
            for _ in range(removed):
                available_connections.put_nowait(None)

    def count_all_num_connections(self, node: ManagedNode) -> int:
        if self.max_connections_per_node:
            return self._created_connections_per_node.get(node.name, 0)

        return sum(i for i in self._created_connections_per_node.values())

    async def get_random_connection(self, primary: bool = False) -> ClusterConnection:
        """Opens new connection to random redis server in the cluster"""
        for node in self.nodes.random_startup_node_iter(primary):
            connection = await self.get_connection_by_node(node)
            if connection:
                return connection
        raise RedisClusterException("Cant reach a single startup node.")

    async def get_connection_by_key(self, key: StringT) -> ClusterConnection:
        if not key:
            raise RedisClusterException(
                "No way to dispatch this command to Redis Cluster."
            )

        return await self.get_connection_by_slot(hash_slot(b(key)))

    async def get_connection_by_slot(self, slot: int) -> ClusterConnection:
        """
        Determines what server a specific slot belongs to and return a redis
        object that is connected
        """
        self.checkpid()

        try:
            return await self.get_connection_by_node(self.get_node_by_slot(slot))
        except KeyError:
            return await self.get_random_connection()

    async def get_connection_by_node(self, node: ManagedNode) -> ClusterConnection:
        """Gets a connection by node"""
        self.checkpid()

        if not self.blocking:
            try:
                connection = self.__node_pool(node.name).get_nowait()
            except asyncio.QueueEmpty:
                connection = None
        else:
            try:
                async with async_timeout.timeout(self.blocking_timeout):
                    connection = await self.__node_pool(node.name).get()
            except asyncio.TimeoutError:
                raise ConnectionError("No connection available.")

        if not connection:
            connection = self._make_node_connection(node)

        self._cluster_in_use_connections.setdefault(node.name, set()).add(connection)
        return cast(ClusterConnection, connection)

    def get_primary_node_by_slot(self, slot: int) -> ManagedNode:
        return self.get_primary_node_by_slots([slot])

    def get_primary_node_by_slots(self, slots: List[int]) -> ManagedNode:
        nodes = {self.nodes.slots[slot][0].node_id for slot in slots}
        if len(nodes) == 1:
            return self.nodes.slots[slots[0]][0]
        else:
            raise RedisClusterException(f"Unable to map slots {slots} to a single node")

    def get_replica_node_by_slot(self, slot: int) -> ManagedNode:
        return self.get_replica_node_by_slots([slot])

    def get_replica_node_by_slots(
        self, slots: List[int], replica_only: bool = False
    ) -> ManagedNode:
        nodes = {self.nodes.slots[slot][0].node_id for slot in slots}
        if len(nodes) == 1:
            slot = slots[0]
            if replica_only:
                return random.choice(
                    [
                        node
                        for node in self.nodes.slots[slot]
                        if node.server_type != "primary"
                    ]
                )
            else:
                return random.choice(self.nodes.slots[slot])
        else:
            raise RedisClusterException(f"Unable to map slots {slots} to a single node")

    def get_node_by_slot(
        self, slot: int, command: Optional[bytes] = None
    ) -> ManagedNode:
        if self.read_from_replicas and command in READONLY_COMMANDS:
            return self.get_replica_node_by_slot(slot)
        return self.get_primary_node_by_slot(slot)

    def get_node_by_slots(
        self, slots: List[int], command: Optional[bytes] = None
    ) -> ManagedNode:
        if self.read_from_replicas and command in READONLY_COMMANDS:
            return self.get_replica_node_by_slots(slots)
        return self.get_primary_node_by_slots(slots)


class BlockingClusterConnectionPool(ClusterConnectionPool):
    """
    .. versionadded:: 4.3.0

    Blocking connection pool for :class:`~coredis.RedisCluster` client

    .. note:: This is just a convenience subclass of :class:`~coredis.pool.ClusterConnectionPool`
       that sets :paramref:`~coredis.pool.ClusterConnectionPool.blocking` to ``True``
    """

    def __init__(
        self,
        startup_nodes: Optional[Iterable[Node]] = None,
        connection_class: Type[ClusterConnection] = ClusterConnection,
        queue_class: Type[asyncio.Queue[Optional[Connection]]] = asyncio.LifoQueue,
        max_connections: Optional[int] = None,
        max_connections_per_node: bool = False,
        reinitialize_steps: Optional[int] = None,
        skip_full_coverage_check: bool = False,
        nodemanager_follow_cluster: bool = True,
        readonly: bool = False,
        read_from_replicas: bool = False,
        max_idle_time: int = 0,
        idle_check_interval: int = 1,
        timeout: int = 20,
        **connection_kwargs: Optional[Any],
    ):
        """

        Changes
          - .. versionchanged:: 4.4.0

            - :paramref:`nodemanager_follow_cluster` now defaults to ``True``

          - .. deprecated:: 4.4.0

            - :paramref:`readonly` renamed to :paramref:`read_from_replicas`

        :param max_connections: Maximum number of connections to allow concurrently from this
         client.
        :param max_connections_per_node: Whether to use the value of :paramref:`max_connections`
         on a per node basis or cluster wide. If ``False`` the per-node connection pools will have
         a maximum size of :paramref:`max_connections` divided by the number of nodes in the
         cluster.
        :param timeout: Number of seconds to block when trying to obtain a connection.
        :param skip_full_coverage_check:
            Skips the check of cluster-require-full-coverage config, useful for clusters
            without the CONFIG command (like aws)
        :param nodemanager_follow_cluster:
            The node manager will during initialization try the last set of nodes that
            it was operating on. This will allow the client to drift along side the cluster
            if the cluster nodes move around alot.
        """
        super().__init__(
            startup_nodes=startup_nodes,
            connection_class=connection_class,
            queue_class=queue_class,
            max_connections=max_connections,
            max_connections_per_node=max_connections_per_node,
            reinitialize_steps=reinitialize_steps,
            skip_full_coverage_check=skip_full_coverage_check,
            nodemanager_follow_cluster=nodemanager_follow_cluster,
            readonly=readonly,
            read_from_replicas=read_from_replicas,
            max_idle_time=max_idle_time,
            idle_check_interval=idle_check_interval,
            timeout=timeout,
            blocking=True,
            **connection_kwargs,
        )



from __future__ import annotations

from coredis.response._callbacks import DictCallback, ResponseCallback
from coredis.typing import (
    AnyStr,
    Dict,
    List,
    Optional,
    ResponsePrimitive,
    Sequence,
    Tuple,
    ValueT,
)


class ACLLogCallback(
    ResponseCallback[
        List[Optional[Sequence[ResponsePrimitive]]],
        List[Optional[Dict[AnyStr, ResponsePrimitive]]],
        Tuple[Optional[Dict[AnyStr, ResponsePrimitive]], ...],
    ]
):
    def transform(
        self,
        response: List[Optional[Sequence[ResponsePrimitive]]],
        **options: Optional[ValueT],
    ) -> Tuple[Optional[Dict[AnyStr, ResponsePrimitive]], ...]:
        return tuple(
            DictCallback[AnyStr, ResponsePrimitive]()(r, version=self.version)
            for r in response
            if r
        )

    def transform_3(
        self,
        response: List[Optional[Dict[AnyStr, ResponsePrimitive]]],
        **options: Optional[ValueT],
    ) -> Tuple[Optional[Dict[AnyStr, ResponsePrimitive]], ...]:
        return tuple(response)
